#include "input.h"
#include "TZR.h"
#include <assert.h>

enum {
	BIND_SCANCODE,
};

struct {
	const char *name;
	struct {
			unsigned int type;
			union {
				SDL_Scancode sc;
			};
	} binds[INPUT_MAX_BINDS];
	int nb_binds;
} input[INPUT_MAX_ACTIONS] = {0};
static int nb_actions = 0;

void
input_new(const char *bind)
{
	assert(nb_actions < INPUT_MAX_ACTIONS);
	input[nb_actions].name = bind;
	nb_actions += 1;
}

void
input_clear(const char *bind)
{
	for (int i = 0; i < nb_actions; i++)
		if (strcmp(bind, input[i].name) == 0) {
			input[i].nb_binds = 0;
			return;
		}
	fprintf(stderr, "input_clear: bind '%s' not found\n", bind);
	assert(0);
}

void
input_bind_scancode(const char *bind, SDL_Scancode sc)
{
	for (int i = 0; i < nb_actions; i++)
		if (strcmp(bind, input[i].name) == 0) {
			assert(input[i].nb_binds < INPUT_MAX_BINDS);
			input[i].binds[input[i].nb_binds].type = BIND_SCANCODE;
			input[i].binds[input[i].nb_binds].sc = sc;
			input[i].nb_binds += 1;
			return;
		}
	fprintf(stderr, "input_bind_scancode: bind '%s' not found\n", bind);
	assert(0);
}

bool
input_down(const char *bind)
{
	for (int i = 0; i < nb_actions; i++) {
		if (strcmp(bind, input[i].name) == 0) {
			for (int k = 0; k < input[i].nb_binds; k++)
				switch (input[i].binds[k].type) {
				case BIND_SCANCODE:
					if (TZR_IsKeyDown(input[i].binds[k].sc))
						return true;
					break;
				default:
					__builtin_unreachable();
				}
			return false;
		}
	}
	fprintf(stderr, "input_down: bind '%s' not found\n", bind);
	assert(0);
}

bool
input_up(const char *bind)
{
	for (int i = 0; i < nb_actions; i++) {
		if (strcmp(bind, input[i].name) == 0) {
			for (int k = 0; k < input[i].nb_binds; k++)
				switch (input[i].binds[k].type) {
				case BIND_SCANCODE:
					if (TZR_IsKeyUp(input[i].binds[k].sc))
						return true;
					break;
				default:
					__builtin_unreachable();
				}
			return false;
		}
	}
	fprintf(stderr, "input_up: bind '%s' not found\n", bind);
	assert(0);
}

bool
input_pressed(const char *bind)
{
	for (int i = 0; i < nb_actions; i++) {
		if (strcmp(bind, input[i].name) == 0) {
			for (int k = 0; k < input[i].nb_binds; k++)
				switch (input[i].binds[k].type) {
				case BIND_SCANCODE:
					if (TZR_IsKeyPressed(input[i].binds[k].sc))
						return true;
					break;
				default:
					__builtin_unreachable();
				}
			return false;
		}
	}
	fprintf(stderr, "input_pressed: bind '%s' not found\n", bind);
	assert(0);
}

bool
input_released(const char *bind)
{
	for (int i = 0; i < nb_actions; i++) {
		if (strcmp(bind, input[i].name) == 0) {
			for (int k = 0; k < input[i].nb_binds; k++)
				switch (input[i].binds[k].type) {
				case BIND_SCANCODE:
					if (TZR_IsKeyReleased(input[i].binds[k].sc))
						return true;
					break;
				default:
					__builtin_unreachable();
				}
			return false;
		}
	}
	fprintf(stderr, "input_released: bind '%s' not found\n", bind);
	assert(0);
}
