#include "TZR.h"
#include "px.h"
#include "input.h"
#include "player.h"
#include "cfg.h"

static int mainloop(void *udata);
static void segment(const PxSpr *spr, int cx, int cy);

int background = 0;

int
main(int argc, char **argv)
{
	(void)argc, (void)argv;

	if (TZR_Init(.width=PX_WIDTH,
	             .height=PX_HEIGHT,
	             .scale=1,
	             .target_fps=TARGET_FPS,
	             .title="can i scream?"))
		return 1;
	if (atexit(TZR_Quit)) {
		perror("main:atexit");
		TZR_Quit();
		return 1;
	}
	if (pxInit())
		return 1;
	if (atexit(pxDeinit)) {
		perror("main:atexit");
		pxDeinit();
		return 1;
	}

	TZR_PlayMusic("res/ost.wav", INT_MAX);

	input_new("move");
	input_bind_scancode("move", SDL_SCANCODE_SPACE);
	input_bind_scancode("move", SDL_SCANCODE_J);
	input_bind_scancode("move", SDL_SCANCODE_F);
	input_bind_scancode("move", SDL_SCANCODE_X);

	Player player = player_init();

	TZR_MainLoop(mainloop, &player);
	return 0;
}

static int
mainloop(void *udata)
{
	extern const PxSpr spr_0;
	extern const PxSpr spr_1;
	extern const PxSpr spr_2;
	extern const PxSpr spr_3;
	extern const PxSpr spr_4;
	extern const PxSpr spr_5;
	extern const PxSpr spr_6;
	static const PxSpr *sprs[] = {
		&spr_0,
		&spr_1,
		&spr_2,
		&spr_3,
		&spr_4,
		&spr_5,
		&spr_6,
	};
	Player *player = udata;

	player_update(player);
	if ((unsigned long)background >= sizeof(sprs)/ sizeof(sprs[0]))
		return 1;

	pxCls(0);
	const double s = 64 * (2 * sin((TZR_GetTick() + rand() % 5 ) / 7.0));
	segment(sprs[background],
	        sprs[background]->w / 2.0 + s * sin(TZR_GetTick() / 32.0),
	        sprs[background]->h / 2.0 + s * cos(TZR_GetTick() / 32.0));
	player_draw(player);

	TZR_DrawBegin();
	pxFlip();
	TZR_DrawEnd();

	return 0;
}

#define SEGRAND (-4 + rand() % 9)

static void
segment(const PxSpr *spr, int cx, int cy)
{
	const int rmargin = spr->w - cx;
	const int bmargin = spr->h - cy;

	pxSpr(spr, -SEGRAND, -SEGRAND, 0, 0, cx, cy);
	pxSpr(spr, cx + SEGRAND, -SEGRAND, cx, 0, rmargin, cy);
	pxSpr(spr, -SEGRAND, cy + SEGRAND, 0, cy, cx, bmargin);
	pxSpr(spr, cx + SEGRAND, cy + SEGRAND, cx, cy, rmargin, bmargin);
}
