/* Licensing information can be found at the end of the file. */
#include "px.h"
#include "TZR.h"
#include <assert.h>

PxCol pxbuf[PX_WIDTH * PX_HEIGHT] = {0};
unsigned int pxzbuf[PX_WIDTH * PX_HEIGHT] = {0};
PxPal pxpal[256] = {
	/* PICO-8 palette */
	{0x00, 0x00, 0x00, false, 0, 0},
	{0x1d, 0x2b, 0x53, false, 0, 0},
	{0x7e, 0x25, 0x53, false, 0, 0},
	{0x00, 0x87, 0x51, false, 0, 0},
	{0xab, 0x52, 0x36, false, 0, 0},
	{0x5f, 0x57, 0x4f, false, 0, 0},
	{0xc2, 0xc3, 0xc7, false, 0, 0},
	{0xff, 0xf1, 0xe8, false, 0, 0},
	{0xff, 0x00, 0x4d, false, 0, 0},
	{0xff, 0xa3, 0x00, false, 0, 0},
	{0xff, 0xec, 0x27, false, 0, 0},
	{0x00, 0xe4, 0x36, false, 0, 0},
	{0x29, 0xad, 0xff, false, 0, 0},
	{0x83, 0x76, 0x9c, false, 0, 0},
	{0xff, 0x77, 0xa8, false, 0, 0},
	{0xff, 0xcc, 0xaa, false, 0, 0},
};
static struct {
	int x;
	int y;
	int w;
	int h;
} pxclip = {0, 0, PX_WIDTH, PX_HEIGHT};
static SDL_Texture *pxtexture = NULL;
static unsigned int pxz = 0;

static unsigned long next = 1;

static int
myrand(void)
{
	next = next * 1103515245 + 12345;
	return((unsigned)(next/65536) % 32768);
}

static void
mysrand(unsigned int seed)
{
	next = seed;
}

int
pxInit(void)
{
	pxtexture = SDL_CreateTexture(___tzr_renderer,
	                              SDL_PIXELFORMAT_ARGB8888,
	                              SDL_TEXTUREACCESS_STREAMING,
	                              PX_WIDTH, PX_HEIGHT);
	if (pxtexture == NULL) {
		fprintf(stderr, "%s\n", SDL_GetError());
		return -1;
	}
	mysrand(rand());
	return 0;
}

void
pxDeinit(void)
{
	if (pxtexture != NULL) {
		SDL_DestroyTexture(pxtexture);
		pxtexture = NULL;
	}
}

int
pxFlip(void)
{
	uint8_t *pixels = NULL;
	int pitch = 0;
	if (SDL_LockTexture(pxtexture, NULL, (void **)&pixels, &pitch)) {
		fprintf(stderr, "%s\n", SDL_GetError());
		return -1;
	}

	for (int y = 0; y < PX_HEIGHT; y++) {
		for (int x = 0; x < PX_WIDTH; x++) {
			const int ix = x << 2;
			PxCol col = pxbuf[x + y * PX_WIDTH];
			if (pxpal[col].spal)
				col = pxpal[col].spal - 1;
			/* random noise */
			col = (col > myrand() % 255) ? 254 : 0;
			pixels[ix + y * pitch + 0] = pxpal[col].b;
			pixels[ix + y * pitch + 1] = pxpal[col].g;
			pixels[ix + y * pitch + 2] = pxpal[col].r;
			pixels[ix + y * pitch + 3] = 255;
		}
	}

	SDL_UnlockTexture(pxtexture);
	SDL_RenderCopy(___tzr_renderer, pxtexture, NULL, NULL);

	return 0;
}

void
pxCls(PxCol c)
{
	for (int y = pxclip.y; y < pxclip.y + pxclip.h; y++)
		for (int x = pxclip.x; x < pxclip.x + pxclip.w; x++) {
			pxbuf[x + y * PX_WIDTH] = c;
			pxzbuf[x + y * PX_WIDTH] = 0;
		}
}

void
pxClip(int x, int y, int w, int h)
{
	pxclip.x = pxClamp(0, PX_WIDTH - 1, x);
	pxclip.y = pxClamp(0, PX_HEIGHT - 1, y);
	pxclip.w = pxClamp(0, PX_WIDTH - pxclip.x - 1, w);
	pxclip.h = pxClamp(0, PX_HEIGHT - pxclip.y - 1, h);
}

void
pxClipReset(void)
{
	pxClip(0, 0, PX_WIDTH, PX_HEIGHT);
}

void
pxZ(unsigned int z)
{
	pxz = z;
}

void
pxPset(int x, int y, PxCol c)
{
	if (x < pxclip.x || x >= pxclip.x + pxclip.w ||
	    y < pxclip.y || y >= pxclip.y + pxclip.h)
		return;

	/* z sorting */
	if (pxz < pxzbuf[x + y * PX_WIDTH])
		return;
	pxzbuf[x + y * PX_WIDTH] = pxz;

	pxbuf[x + y * PX_WIDTH] = pxpal[c].pal ? (pxpal[c].pal - 1) : c;
}

uint8_t
pxPget(int x, int y)
{
	if (x < 0 || y < 0 || x >= PX_WIDTH || y >= PX_HEIGHT)
		return 0;
	return pxbuf[x + y * PX_WIDTH];
}

void
pxRect(int x0, int y0, int x1, int y1, PxCol col)
{
	if (x0 > x1) {
		x0 ^= x1;
		x1 ^= x0;
		x0 ^= x1;
	}
	if (y0 > y1) {
		y0 ^= y1;
		y1 ^= y0;
		y0 ^= y1;
	}
	for (int x = x0; x <= x1; x++) {
		pxPset(x, y0, col);
		pxPset(x, y1, col);
	}
	for (int y = y0; y <= y1; y++) {
		pxPset(x0, y, col);
		pxPset(x1, y, col);
	}
}

void
pxRectfill(int x0, int y0, int x1, int y1, PxCol col)
{
	if (x0 > x1) {
		x0 ^= x1;
		x1 ^= x0;
		x0 ^= x1;
	}
	if (y0 > y1) {
		y0 ^= y1;
		y1 ^= y0;
		y0 ^= y1;
	}
	for (int y = y0; y <= y1; y++)
		for (int x = x0; x <= x1; x++)
			pxPset(x, y, col);
}

void
pxLine(int x0, int y0, int x1, int y1, PxCol col)
{
	int dx, dy, sx, sy, err, e2;

	dx = x1 - x0;
	dx = (dx < 0) ? (-dx) : (dx);
	dy = y1 - y0;
	dy = (dy < 0) ? (dy) : (-dy);
	sx = (x0 < x1) ? (1) : (-1);
	sy = (y0 < y1) ? (1) : (-1);
	err = dx + dy;

	for (;;) {
		pxPset(x0, y0, col);

		if (x0 == x1 && y0 == y1)
			break;

		e2 = 2 * err;
		if (e2 >= dy) {
			err += dy;
			x0 += sx;
		}
		if (e2 <= dx) {
			err += dx;
			y0 += sy;
		}
	}
}

void
_pxSpr(const PxSprArgs *args)
{
	int w = (args->w < 0) ? (int)(args->spr->w) : (args->w);
	w = pxMin(w, args->spr->w - args->ix);

	int h = (args->h < 0) ? (int)(args->spr->h) : (args->h);
	h = pxMin(h, args->spr->h - args->iy);

	for (int y = 0; y < h; y++) {
		for (int x = 0; x < w; x++) {
			const int dx = x + args->x;
			const int dy = y + args->y;
			int ix = args->ix + ((args->flip_x)
			                     ? (w - x - 1)
			                     : (x));
			int iy = args->iy + ((args->flip_y)
			                     ? (h - y - 1)
			                     : (y));
			const PxCol c = args->spr->data[ix + iy * args->spr->w];
			if (!pxpal[c].t)
				pxPset(dx, dy, c);
		}
	}
}

void
_pxPal(const PxPalArgs *args)
{
	assert((args->c0 == -1) ^ (args->c1 != -1));
	if (args->c0 == -1)
		for (int i = 0; i < 256; i++)
			pxpal[i].pal = 0;
	else
		pxpal[args->c0].pal = args->c1 + 1;
}

void
_pxSpal(const PxSpalArgs *args)
{
	assert((args->c0 == -1) ^ (args->c1 != -1));
	if (args->c0 == -1)
		for (int i = 0; i < 256; i++)
			pxpal[i].spal = 0;
	else
		pxpal[args->c0].spal = args->c1 + 1;
}

void
_pxPalt(const PxPaltArgs *args)
{
	if (args->col < 0 || args->col > 255)
		for (int i = 0; i < 256; i++)
			pxpal[i].t = false;
	else
		pxpal[args->col].t = args->t;
}

int
pxMin(int a, int b)
{
	return (a < b) ? a : b;
}

int
pxMax(int a, int b)
{
	return (a > b) ? a : b;
}

int
pxClamp(int min, int max, int v)
{
	return pxMax(min, pxMin(max, v));
}

/*
** Copyright (c) 2023 kdx
**
** Permission is hereby granted, free of charge, to any person obtaining a copy
** of this software and associated documentation files (the "Software"), to
** deal in the Software without restriction, including without limitation the
** rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
** sell copies of the Software, and to permit persons to whom the Software is
** furnished to do so, subject to the following conditions:
**
** The above copyright notice and this permission notice shall be included in
** all copies or substantial portions of the Software.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
** FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
** AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
** LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
** FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
** IN THE SOFTWARE.
*/
