/* Licensing information can be found at the end of the file. */
#pragma once
#include <stdint.h>
#include <stdbool.h>

#ifndef PX_WIDTH
#define PX_WIDTH 128
#endif
#ifndef PX_HEIGHT
#define PX_HEIGHT 128
#endif

typedef uint8_t PxCol;

typedef struct {
	uint8_t r, g, b;
	bool t; /* tranparency */
	int pal, spal;
} PxPal;

typedef struct {
	unsigned int w, h;
	PxCol *data;
} PxSpr;

typedef struct {
	int _;
	const PxSpr *spr;
	int x;
	int y;
	int ix;
	int iy;
	int w;
	int h;
	bool flip_x;
	bool flip_y;
} PxSprArgs;

typedef struct {
	int _;
	int c0;
	int c1;
} PxPalArgs;
typedef PxPalArgs PxSpalArgs;

typedef struct {
	int _;
	int col;
	bool t;
} PxPaltArgs;

extern PxCol pxbuf[PX_WIDTH * PX_HEIGHT];
extern PxPal pxpal[256];

/* Return -1 on error.
 * Should be called after TZR_Init. */
int pxInit(void);

/* Should be called before TZR_Quit. */
void pxDeinit(void);

/* Return -1 on error.
 * Should be called between TZR_DrawBegin and TZR_DrawEnd. */
int pxFlip(void);

/*** DRAW ***/
/* Fill and reset Z on the clipping region with color 'c'. */
void pxCls(PxCol c);

/* Set the clipping region. */
void pxClip(int x, int y, int w, int h);

/* Reset the clipping region. */
void pxClipReset(void);

/* Set Z layer. */
void pxZ(unsigned int z);

/* Put a pixel. */
void pxPset(int x, int y, PxCol c);

/* Get a pixel. */
PxCol pxPget(int x, int y);

/* Draw empty rectangle. */
void pxRect(int x0, int y0, int x1, int y1, PxCol col);

/* Draw filled rectangle. */
void pxRectfill(int x0, int y0, int x1, int y1, PxCol col);

/* Draw line. */
void pxLine(int x0, int y0, int x1, int y1, PxCol col);

/* Draw sprite. */
#define pxSpr(...) _pxSpr(&(const PxSprArgs){ \
	.x=0, .y=0, .ix=0, .iy=0, .w=-1, .h=-1, .flip_x=false, .flip_x=false, \
	._=0, __VA_ARGS__})
void _pxSpr(const PxSprArgs *args);

/* Bind palette index to new color. */
#define pxPal(...) _pxPal(&(const PxPalArgs){ \
	.c0=-1, .c1=-1, ._=0, __VA_ARGS__})
void _pxPal(const PxPalArgs *args);

/* Set screen palette. */
#define pxSpal(...) _pxSpal(&(const PxSpalArgs){ \
	.c0=-1, .c1=-1, ._=0, __VA_ARGS__})
void _pxSpal(const PxSpalArgs *args);

/* Set palette tranparency. */
#define pxPalt(...) _pxPalt(&(const PxPaltArgs){ \
	.col=-1, .t=true, ._=0, __VA_ARGS__})
void _pxPalt(const PxPaltArgs *args);

/*** MATH ***/
int pxMin(int a, int b);
int pxMax(int a, int b);
int pxClamp(int min, int max, int v);

/*
** Copyright (c) 2023 kdx
**
** Permission is hereby granted, free of charge, to any person obtaining a copy
** of this software and associated documentation files (the "Software"), to
** deal in the Software without restriction, including without limitation the
** rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
** sell copies of the Software, and to permit persons to whom the Software is
** furnished to do so, subject to the following conditions:
**
** The above copyright notice and this permission notice shall be included in
** all copies or substantial portions of the Software.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
** FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
** AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
** LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
** FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
** IN THE SOFTWARE.
*/
